/*
Copyright (c) 2000-2005 Joachim Henke

For conditions of distribution and use, see copyright notice in base91.h
*/

struct bit_queue {
	unsigned long d;
	char n;	/* number of bits in queue */
};

const unsigned char enctab[91] = {
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
	'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '!', '#', '$',
	'%', '&', '(', ')', '*', '+', ',', '.', '/', ':', ';', '<', '=',
	'>', '?', '@', '[', ']', '^', '_', '`', '{', '|', '}', '~', '"'
};
char dectab[256];
struct bit_queue ebq, dbq;
int dv;

void b91_enc_reset(void)
{
	ebq.d = 0;
	ebq.n = 0;
}

void b91_dec_reset(void)
{
	dbq.d = 0;
	dbq.n = 0;
	dv = -1;
}

void b91_init(void)
{
	int i;

	for (i = 0; i < 256; ++i)	/* build decoding table */
		dectab[i] = -1;
	for (i = 0; i < 91; ++i)
		dectab[enctab[i]] = i;
	b91_enc_reset();
	b91_dec_reset();
}

int b91_encode(unsigned char *ib, int l, unsigned char *ob)
{
	int v, i, o = 0;

	for (i = 0; i < l; ++i) {
		ebq.d |= ib[i] << ebq.n;
		ebq.n += 8;
		if (ebq.n > 13) {	/* enough bits in queue */
			v = ebq.d & 8191;
			if (v > 88) {
				ebq.d >>= 13;
				ebq.n -= 13;
			} else {	/* we can take 14 bits */
				v = ebq.d & 16383;
				ebq.d >>= 14;
				ebq.n -= 14;
			}
			ob[o++] = enctab[v % 91];
			ob[o++] = enctab[v / 91];
		}
	}

	return o;
}

int b91_enc_end(unsigned char *ob)
{
	int o = 0;

	if (ebq.n) {	/* write remaining bits from bit queue */
		ob[o++] = enctab[ebq.d % 91];
		if (ebq.n > 7 || ebq.d > 90)
			ob[o++] = enctab[ebq.d / 91];
	}
	b91_enc_reset();

	return o;
}

int b91_decode(unsigned char *ib, int l, unsigned char *ob)
{
	int i, o = 0;

	for (i = 0; i < l; ++i) {
		if (dectab[ib[i]] < 0)	/* ignore invalid chars */
			continue;
		if (dv < 0)	/* start next value */
			dv = dectab[ib[i]];
		else {
			dv += dectab[ib[i]] * 91;
			dbq.d |= dv << dbq.n;
			dbq.n += (dv & 8191) > 88 ? 13 : 14;
			do {
				ob[o++] = (unsigned char) dbq.d;
				dbq.d >>= 8;
				dbq.n -= 8;
			} while (dbq.n > 7);
			dv = -1;	/* mark value complete */
		}
	}

	return o;
}

int b91_dec_end(unsigned char *ob)
{
	int o = 0;

	if (dv + 1)	/* write remaining bits */
		ob[o++] = (unsigned char) (dbq.d | dv << dbq.n);
	b91_dec_reset();

	return o;
}
